/*
    ZEsarUX  ZX Second-Emulator And Released for UniX 
    Copyright (C) 2013 Cesar Hernandez Bano

    This file is part of ZEsarUX.

    ZEsarUX is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <stdio.h>
#include <stdlib.h>
#include <dirent.h>


#include "divmmc.h"
#include "mmc.h"
#include "cpu.h"
#include "debug.h"
#include "utils.h"
#include "menu.h"
#include "screen.h"
#include "operaciones.h"
#include "zxuno.h"
#include "contend.h"
#include "prism.h"

z80_bit divmmc_enabled={0};


z80_byte *divmmc_memory_pointer;

#define DIVMMC_FIRMWARE_KB 8
#define DIVMMC_FIRMWARE_ALLOCATED_KB 64
#define DIVMMC_RAM_ALLOCATED_KB 128
#define DIVMMC_ROM_NAME "esxmmc085.rom"

//puntero a cpu core normal sin divmmc
//void (*cpu_core_loop_no_divmmc) (void);


z80_byte divmmc_control_register;

//Si esta la paginacion de divmmc activa
z80_bit divmmc_paginacion_activa={0};

        int divmmc_salta_trap_antes=0;
        int divmmc_salta_trap_despues=0;
	int divmmc_salta_trap_despaginacion_despues=0;	


//Core de cpu loop para hacer traps de cpu
void divmmc_pre_opcode_fetch(void)
{

/*
Memory mapping could be invoked manually (by setting CONMEM), or automatically
(CPU has fetched opcode form an entry-point). Automatic mapping is active
only if EPROM/EEPROM is present (jumper EPROM is closed) or bit MAPRAM is set.
Automatic mapping occurs at the begining of refresh cycle after fetching
opcodes (M1 cycle) from 0000h, 0008h, 0038h, 0066h, 04c6h and 0562h. It's
also mapped instantly (100ns after /MREQ of that fetch is falling down) after
executing opcode from area 3d00..3dffh. Memory is automatically disconnected in
refresh cycle of the instruction fetch from so called off-area, which is 
1ff8-1fffh.
*/



	divmmc_salta_trap_antes=0;
	divmmc_salta_trap_despues=0;
	divmmc_salta_trap_despaginacion_despues=0;

	//Si en ZXUNO y DIVEN desactivado
	if (MACHINE_IS_ZXUNO_DIVEN_DISABLED) return;

	//Si en Prism y rom 0, no saltar
	if (MACHINE_IS_PRISM) {
		if (prism_rom_page==0) {
			//printf ("rom 0 en prism, no saltar divmmc. pc=%d\n",reg_pc);
			return;
		}
	}

	//Traps que paginan memoria y saltan despues de leer instruccion
	switch (reg_pc) {
		case 0x0000:
		case 0x0008:
		case 0x0038:
		case 0x0066:
		case 0x04c6:
		case 0x0562:
			if (divmmc_paginacion_activa.v==0) divmmc_salta_trap_despues=1;
		break;
	}


	//Traps que paginan memoria y saltan antes de leer instruccion
	if (reg_pc>=0x3d00 && reg_pc<=0x3dff) divmmc_salta_trap_antes=1;

	//Traps que despaginan memoria antes de leer instruccion
	if (reg_pc>=0x1ff8 && reg_pc<=0x1fff && divmmc_paginacion_activa.v) {
		//printf ("Saltado trap de despaginacion pc actual: %d\n",reg_pc);
		divmmc_salta_trap_despaginacion_despues=1;
        }
	


	if (divmmc_salta_trap_antes && divmmc_paginacion_activa.v==0) {
		//printf ("Saltado trap de paginacion antes pc actual: %d\n",reg_pc);
		divmmc_paginacion_activa.v=1;
        }


}


void divmmc_post_opcode_fetch(void)
{

	if (divmmc_salta_trap_despues) {
		//printf ("Saltado trap de paginacion despues. pc actual: %d\n",reg_pc);
		divmmc_paginacion_activa.v=1;
	}

	if (divmmc_salta_trap_despaginacion_despues) {
		//printf ("Saltado trap de despaginacion despues. pc actual: %d\n",reg_pc);
		divmmc_paginacion_activa.v=0;
	}


}


void divmmc_write_control_register(z80_byte value)
{

	//printf ("Escribiendo registro de control divmmc valor: 0x%02X\n",value);
	if (value&128) {
		//printf ("Activando paginacion\n");
		divmmc_paginacion_activa.v=1;
	}
	else {
		if ((value&63)==0) {
			//printf ("Desactivando paginacion\n");
			//divmmc_paginacion_activa.v=0;
		}
	}

	//TODO. Bit MAPRAM can be set to '1' only (when it's '1', only power-on can return it to '0'). 
	divmmc_paginacion_activa.v=1;


	divmmc_control_register=value;
}
	

//Rutinas originales antes de cambiarlas
void (*divmmc_original_poke_byte)(z80_int dir,z80_byte valor);
void (*divmmc_original_poke_byte_no_time)(z80_int dir,z80_byte valor);
z80_byte (*divmmc_original_peek_byte)(z80_int dir);
z80_byte (*divmmc_original_peek_byte_no_time)(z80_int dir);


//Retorna puntero a direccion apuntada por dir, en caso de paginacion divmmc activa y dir <4000h
z80_byte *divmmc_return_memory_paged_pointer(z80_int dir)
{
/*


    7        6     5  4  3  2   1       0
[ CONMEM , MAPRAM, X, X, X, X, BANK1, BANK0 ]

So, when CONMEM is set, there is:
0000-1fffh - EEPROM/EPROM/NOTHING(if empty socket), and this area is
	     flash-writable if EPROM jumper is open.
2000-3fffh - 8k bank, selected by BANK 0..1 bits, always writable.

When MAPRAM is set, but CONMEM is zero, and entrypoint was reached:
0000-1fffh - Bank No.3, read-only
2000-3fffh - 8k bank, selected by BANK 0..1. If it's different from Bank No.3,
	     it's writable.

When MAPRAM is zero, CONMEM is zero, EPROM jumper is closed and entrypoint was
reached:
0000-1fffh - EEPROM/EPROM/NOTHING(if empty socket, so open jumper in this case),
	     read-only.
2000-3fffh - 8k bank, selected by BANK 0..1, always writable.
*/

	
	//De momento facil
	if (dir<=0x1fff) {
		if (divmmc_control_register&64) {
			//printf ("Retornando direccion cuando MAPRAM activo\n");
			//pagina ram 3.
			int pagina=3;
			int offset=DIVMMC_FIRMWARE_ALLOCATED_KB*1024;
			offset +=pagina*8192;
			return &divmmc_memory_pointer[offset];
		}

		else {
			//Devolver eprom
			return &divmmc_memory_pointer[dir];
		}
	}

	else {
		//de 2000h a 3fffh
		//Devolver pagina de memoria ram
		int pagina=divmmc_control_register&15;

		//Saltar los primeros kb asignados para firmware
		int offset=DIVMMC_FIRMWARE_ALLOCATED_KB*1024;

		//Nos ubicamos en la pagina
		offset +=pagina*8192;

		//Nos ubicamos en la direccion
		offset +=(dir-0x2000);
		return &divmmc_memory_pointer[offset];
	}
		
}

//Rutinas propias

//Comentamos estas y las simplificamos a continuacion, ademas sin tener que sumar t_estados fijos
/*
void divmmc_poke_byte_no_time(z80_int dir,z80_byte valor)
{
        if (divmmc_paginacion_activa.v==0) {
                divmmc_original_poke_byte_no_time(dir,valor);
		return;
	}
	else {
		//Si poke a eprom. De momento no permitir modo MAPRAM
		if (dir<8192) return;

		if (dir<16384) {
			z80_byte *puntero=divmmc_return_memory_paged_pointer(dir);
			*puntero=valor;
		}
		else divmmc_original_poke_byte_no_time(dir,valor);
	}
}

void divmmc_poke_byte(z80_int dir,z80_byte valor)
{
	//T_estados y contend como en 48k
#ifdef EMULATE_CONTEND
        if ( (dir&49152)==16384) {
                t_estados += contend_table[ t_estados ];
        }
#endif

        //Y sumamos estados normales
        t_estados += 3;


        if (dir>16383) {
#ifdef EMULATE_VISUALMEM

set_visualmembuffer(dir);

#endif
	}



	divmmc_poke_byte_no_time(dir,valor);
}
*/

//Si hay que hacer poke a memoria interna
void divmmc_poke_byte_to_internal_memory(z80_int dir,z80_byte valor)
{       
        if (divmmc_paginacion_activa.v==0) {
                return;
        }
        else {  
                //Si poke a eprom. De momento no permitir modo MAPRAM
                if (dir<8192) return;
                
                if (dir<16384) { 
                        z80_byte *puntero=divmmc_return_memory_paged_pointer(dir);
                        *puntero=valor;
                }
        }
}

void divmmc_poke_byte_no_time(z80_int dir,z80_byte valor)
{
	divmmc_poke_byte_to_internal_memory(dir,valor);
	//Asumimos que divmmc_poke_byte_to_internal_memory puede hacer poke a la memoria interna de divmmc,
	//y luego divmmc_original_poke_byte_no_time hara poke a memoria normal si conviene
	//Realmente no se dara el caso dado que la memoria interna de divmmc se mapea en 0-16383
	//y en esas direcciones luego divmmc_original_poke_byte_no_time es un poke a rom que no esta permitido
	divmmc_original_poke_byte_no_time(dir,valor);
}

void divmmc_poke_byte(z80_int dir,z80_byte valor)
{
	divmmc_poke_byte_to_internal_memory(dir,valor);
	//Asumimos que divmmc_poke_byte_to_internal_memory puede hacer poke a la memoria interna de divmmc,
	//y luego divmmc_original_poke_byte hara poke a memoria normal si conviene
	//Realmente no se dara el caso dado que la memoria interna de divmmc se mapea en 0-16383
	//y en esas direcciones luego divmmc_original_poke_byte_no_time es un poke a rom que no esta permitido
	divmmc_original_poke_byte(dir,valor);
}




//TODO: simplificar esto sin tener que sumar t_estados desde aqui a mano
//quiza no es tan facil como las rutinas de poke dado que aqui el resultado de peek depende de si esta paginada la rom normal o la ram divmmc
z80_byte divmmc_peek_byte_no_time(z80_int dir)
{
	if (divmmc_paginacion_activa.v==0) {
                return divmmc_original_peek_byte_no_time(dir);
	}
	else {
                if (dir<16384) {
                        z80_byte *puntero=divmmc_return_memory_paged_pointer(dir);
                        return *puntero;
                }
                else return divmmc_original_peek_byte_no_time(dir);

	}
}

z80_byte divmmc_peek_byte(z80_int dir)
{
        //T_estados y contend como en 48k
#ifdef EMULATE_CONTEND
        if ( (dir&49152)==16384) {
                t_estados += contend_table[ t_estados ];
        }
#endif

        //Y sumamos estados normales
        t_estados += 3;

	return divmmc_peek_byte_no_time(dir);
}

//Establecer rutinas propias
void divmmc_set_peek_poke_functions(void)
{
	//Ver si ya no estaban activas
	if (poke_byte!=divmmc_poke_byte) {
		debug_printf (VERBOSE_DEBUG,"Setting DIVMMC poke / peek functions");
		//Guardar anteriores
		divmmc_original_poke_byte=poke_byte;
		divmmc_original_poke_byte_no_time=poke_byte_no_time;
		divmmc_original_peek_byte=peek_byte;
		divmmc_original_peek_byte_no_time=peek_byte_no_time;

		//Modificar y poner las de divmmc
		poke_byte=divmmc_poke_byte;
		poke_byte_no_time=divmmc_poke_byte_no_time;
		peek_byte=divmmc_peek_byte;
		peek_byte_no_time=divmmc_peek_byte_no_time;
	}
}

//Restaurar rutinas de divmmc
void divmmc_restore_peek_poke_functions(void)
{       
		debug_printf (VERBOSE_DEBUG,"Restoring original poke / peek functions before DIVMMC");
                poke_byte=divmmc_original_poke_byte;
                poke_byte_no_time=divmmc_original_poke_byte_no_time;
                peek_byte=divmmc_original_peek_byte;
                peek_byte_no_time=divmmc_original_peek_byte_no_time;
}


//Retorna 0 si ok
int divmmc_load_firmware(void)
{
        FILE *ptr_divmmc_romfile;
        int leidos=0;

	debug_printf (VERBOSE_INFO,"Loading divmmc firmware %s",DIVMMC_ROM_NAME);
        
	ptr_divmmc_romfile=fopen(DIVMMC_ROM_NAME,"rb");
        
        
        if (ptr_divmmc_romfile!=NULL) {
                
                leidos=fread(divmmc_memory_pointer,1,DIVMMC_FIRMWARE_KB*1024,ptr_divmmc_romfile);
                fclose(ptr_divmmc_romfile);
                
        }
        
        
        
        if (leidos!=DIVMMC_FIRMWARE_KB*1024 || ptr_divmmc_romfile==NULL) {
                debug_printf (VERBOSE_ERR,"Error reading DIVMMC firmware");
		//Lo desactivamos asi porque el disable hace otras cosas, como cambiar el core loop, que no queremos
		divmmc_enabled.v=0;
		return 1;
        }

	return 0;
}






void divmmc_alloc_memory(void)
{
	int size=(DIVMMC_FIRMWARE_ALLOCATED_KB+DIVMMC_RAM_ALLOCATED_KB)*1024;

	debug_printf (VERBOSE_DEBUG,"Allocating %d kb of memory for divmmc emulation",size/1024);

        divmmc_memory_pointer=malloc(size);
        if (divmmc_memory_pointer==NULL) {
                cpu_panic ("No enough memory for divmmc emulation");
        }

}

void divmmc_enable(void)
{

	debug_printf (VERBOSE_INFO,"Enabling divmmc");

	//Asignar memoria, si es que no estamos en modo ZXUNO
	//Cargar firmware, si es que no estamos en modo ZXUNO

	if (MACHINE_IS_ZXUNO) {
		//Pagina 12 del ZX-Uno
		divmmc_memory_pointer=zxuno_sram_mem_table[12];

	}

	else {

		//Memoria asignada para divmmc:
		//Primeros 64kb: solo se usan 8kb para el firmware
		//Siguientes 128kb: paginas ram para divmmc
		//Este modelo de memoria es compatible con zxuno

		divmmc_alloc_memory();

		if (divmmc_load_firmware()) return;

	}


	//Cambiar poke, peek funciones
	divmmc_set_peek_poke_functions();


	divmmc_control_register=0;
	divmmc_paginacion_activa.v=0;
	divmmc_enabled.v=1;

}

void divmmc_disable(void)
{

	if (divmmc_enabled.v==0) return;

	debug_printf (VERBOSE_INFO,"Disabling divmmc");


	//Restaurar poke, peek funciones
	divmmc_restore_peek_poke_functions();


        divmmc_enabled.v=0;
}
