/*
    ZEsarUX  ZX Second-Emulator And Released for UniX 
    Copyright (C) 2013 Cesar Hernandez Bano

    This file is part of ZEsarUX.

    ZEsarUX is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <stdio.h>
#include <string.h>


#include "ulaplus.h"
#include "screen.h"
#include "debug.h"
#include "spectra.h"
#include "menu.h"



/*
Ulaplus incompatible con scanlines o gigascreen
Ula plus al cambiar un color, clear putpixel
Ula plus tabla de conversión de valor r,g de 3 bits a 8 bits. Tabla de valor b de 2 bits a 8 bits

Ula plus http://scratchpad.wikia.com/wiki/ZX_Spectrum_64_Colour_RGB_Mode

http://scratchpad.wikia.com/wiki/ULAplus

https://sites.google.com/site/ulaplus/home


*/

//Indica si esta presente el soporte de ulaplus
z80_bit ulaplus_presente={0};



//Indica si esta activo ulaplus
z80_bit ulaplus_enabled={0};

//Ultimo valor enviado a puerto BF3B
z80_byte ulaplus_last_send_BF3B;

//Ultimo valor enviado a puerto FF3B
z80_byte ulaplus_last_send_FF3B;

//modo ulaplus actual. modo 0 es desactivado (paleta normal spectrum), modo 1 rgb, modo 3 radastan
z80_byte ulaplus_mode=0;

//Paleta actual de ulaplus
z80_byte ulaplus_palette_table[64];

/*
How do you switch on the 64-colour mode?
OUT 48955,64: OUT 65339,1.

How do you switch off the 64-colour mode?
OUT 48955,64: OUT 65339,0.
*/

/*

0xBF3B is the register port (write only)

The byte output will be interpreted as follows:

Bits 0-5: Select the register sub-group
Bits 6-7: Select the register group. Two groups are currently available:
     00 - palette group
          When this group is selected, the sub-group determines the
          entry in the palette table (0-63).
     01 - mode group
          The sub-group must be zero. Sub-group entries 1-63 are
          reserved for mode specific controls.

0xFF3B is the data (read/write)

When the palette group is selected, the byte output will be interpreted as follows:

Bits 0-1: Blue intensity.
Bits 2-4: Red intensity.
Bits 5-7: Green intensity.

*/




//tabla de conversion de colores de rgb ulaplus (8 bit) a 32 bit
//Cada elemento del array contiene el valor rgb real, por ejemplo,
//un valor rgb 33 de ulaplus, en la posicion 33 de este array retorna el color en formato rgb de 32 bits
int ulaplus_rgb_table[256];
/*
Colores 8 bit rgb de ulaplus:
Bits 0-1: Blue intensity.
Bits 2-4: Red intensity.
Bits 5-7: Green intensity.

This mode uses a sub-set of 9-bit RGB. The low bit is duplicated (Bb becomes Bbb). This gives access to a fixed half the potential 512 colour palette. The jump in intensity happens in the lower range where it is less noticeable. 

La tabla ulaplus_rgb_table tiene para cada valor posible de rgb de ula plus su correspondiente en 24 bits (8 bit r, 8 bit g, 8 bit b)


*/

//Inicializar tabla de colores de rgb ulaplus (8 bit) a 32 bit
void init_ulaplus_table(void)
{
	debug_printf (VERBOSE_DEBUG,"Initializing ULAplus rgb color table conversion");
	//Cada componente r,g,b de ulaplus se pasa a 8 bits, sumando tendremos color de 24 bits
	//Componentes red y green son de 3 bits, y b es de 2 bits
	//Maximo r,g: 3 bits a 1 -> 7. maximo en 8 bits: 255 . 255/7=36.42
	//Valor para 2 por ejemplo = 2 * 36.42 = 72.85
	
	//Minitablas de conversion de 3 bits a 8 bits
	z80_byte color_3_to_8[8]={
	0,36,73,109,146,182,219,255
	};

	int color,color32;
	z80_byte r,g,b;
	z80_byte r8,g8,b8;

	for (color=0;color<256;color++) {
		r=(color>>2)&7;
		g=(color>>5)&7;

		//componente b es un tanto esoterico
		//The missing lowest blue bit is set to OR of the other two blue bits (Bb becomes 000 for 00, and Bb1 for anything else)
		b=(color&3);
		b=(b<<1);
		if (b) b=b|1;

		//Pasamos cada componente de 3 bits a su correspondiente de 8 bits
		r8=color_3_to_8[r];
		g8=color_3_to_8[g];
		b8=color_3_to_8[b];

		color32=(r8<<16)|(g8<<8)|b8;
		ulaplus_rgb_table[color]=color32;

		debug_printf (VERBOSE_DEBUG,"ULAplus RGB 0x%02X is 0x%06X 32 bit RGB",color,color32);
	}

	debug_printf (VERBOSE_DEBUG,"Initializing ULAplus 64 colour table to black");
	int i;
	for (i=0;i<64;i++) ulaplus_palette_table[i]=0;
	
}



//Cambiar un color de la paleta ulaplus
//Entrada: indice entre 0...63 y color formato rgb8 de ulaplus
void ulaplus_change_palette_colour(z80_byte index,z80_byte color)
{
	//Actualizamos entrada en tabla ulaplus
	ulaplus_palette_table[index]=color;

}


void disable_ulaplus(void)
{
	debug_printf (VERBOSE_INFO,"Disabling ULAplus");
	ulaplus_presente.v=0;
}

void enable_ulaplus(void)
{
        debug_printf (VERBOSE_INFO,"Enabling ULAplus");
        ulaplus_presente.v=1;

        //son excluyentes
        //disable_interlace();
	disable_scanlines();
	disable_gigascreen();
	spectra_disable();

        //necesita real video
        enable_rainbow();
}


void ulaplus_set_mode(z80_byte value)
{
                                //establecer modo
                                //modo 0: normal spectrum->ulaplus desactivado. modos 1 rgb. modos 2 y superiores no implementados
                                /*

                                modos 3 y superiores usan 4 bits por pixel. van en contra de la ultima especificacion,
                                donde dice que solo se usa bit 0 y bit 1, y bit 1 indica escala de grises

                                modo 3: radastan 128x96

                                modo 5: 256x96

                                modo 7: 128x192

                                modo 9: 256x192. no mira ram 5 o 7, sino que lee siempre a partir de la direccion 16384 en adelante
                                */

                                z80_byte ulaplus_mode_anterior=ulaplus_mode;

                                ulaplus_mode=(value&63);

                                switch (ulaplus_mode) {
                                        case 0:
                                                debug_printf (VERBOSE_DEBUG,"Disabling ULAplus (mode 0)");
                                                ulaplus_enabled.v=0;
                                                if (ulaplus_mode!=ulaplus_mode_anterior) {
                                                        screen_print_splash_text(10,0,7+8,"Disabling ULAplus (mode 0)");
                                                }

                                        break;

                                        case 1:
                                                ulaplus_enabled.v=1;
                                                debug_printf (VERBOSE_DEBUG,"Enabling ULAplus mode 1. RGB");
                                                if (ulaplus_mode!=ulaplus_mode_anterior) {
                                                        screen_print_splash_text(10,0,7+8,"Enabling ULAplus mode 1. RGB");
                                                }
                                        break;

                                        case 3:
                                                ulaplus_enabled.v=1;
					        debug_printf (VERBOSE_DEBUG,"Enabling ULAplus linear mode 3. Radastan. 128x96");
                                                if (ulaplus_mode!=ulaplus_mode_anterior) {
                                                        screen_print_splash_text(10,0,7+8,"Enabling ULAplus linear mode 3. Radastan. 128x96");
                                                }

                                        break;

                                        case 5:
                                                ulaplus_enabled.v=1;
                                                debug_printf (VERBOSE_DEBUG,"Enabling ULAplus linear mode 5. 256x96");
                                                if (ulaplus_mode!=ulaplus_mode_anterior) {
                                                        screen_print_splash_text(10,0,7+8,"Enabling ULAplus linear mode 5. 256x96");
                                                }

                                        break;

                                        case 7:
                                                ulaplus_enabled.v=1;
                                                debug_printf (VERBOSE_DEBUG,"Enabling ULAplus linear mode 7. 128x192");
                                                if (ulaplus_mode!=ulaplus_mode_anterior) {
                                                        screen_print_splash_text(10,0,7+8,"Enabling ULAplus linear mode 7. 128x192");
                                                }

                                        break;

                                        case 9:
                                                ulaplus_enabled.v=1;
                                                debug_printf (VERBOSE_DEBUG,"Enabling ULAplus linear mode 9. 256x192");
                                                if (ulaplus_mode!=ulaplus_mode_anterior) {
                                                        screen_print_splash_text(10,0,7+8,"Enabling ULAplus linear mode 9. 256x192");
                                                }

                                        break;


                                        default:
                                                debug_printf (VERBOSE_DEBUG,"Unknown ulaplus mode %d",ulaplus_mode);
                                        break;

                                }
}
